#include "generator.h"
#include "xgegrammar.h"
#include "nicai/xml/stream.h"
#include "nicai/each.h"
#include <fstream>
#include <algorithm>


class generator_impl
{
  public:
    std::string basename;
    xgegrammar::XgeDocument * doc;
    std::ofstream cpp;
    std::ofstream hpp;

    std::string active_cpp_namespace;
    std::string active_hpp_namespace;

    std::map<std::string, std::string> class_names;

    std::string transform_type(const std::string & atype, const std::string & def="");

    void generate_code();
    void generate_code(xgegrammar::Grammar & node);

    void generate_class_declaration(xgegrammar::Grammar & node);
    void generate_class_definition(xgegrammar::Grammar & node);
    void generate_class_implementation(xgegrammar::Grammar & node);

    void generate_class_declaration(xgegrammar::ComplexClass & node);
    void generate_class_definition(xgegrammar::ComplexClass & node);
    void generate_class_implementation(xgegrammar::ComplexClass & node);

    void generate_class_declaration(xgegrammar::DocumentRoot & node);
    void generate_class_definition(xgegrammar::DocumentRoot & node);
    void generate_class_implementation(xgegrammar::DocumentRoot & node);

    void generate_class_definition(xgegrammar::Attribute & node);
    void generate_class_implementation(xgegrammar::Attribute & node);

    void generate_class_implementation(xgegrammar::Element & node);

    ~generator_impl() {}
};

////////////////////////////////////////////////////////////////////////////////
// helper functions
////////////////////////////////////////////////////////////////////////////////

template <class T>
class each_child : public nicai::each<typename nicai::xml::poly_childs<T>::iterator>
{
  public:
    each_child(nicai::xml::element & node)
    : nicai::each<typename nicai::xml::poly_childs<T>::iterator> (
          nicai::xml::poly_childs<T>(node.getChilds()).begin(),
          nicai::xml::poly_childs<T>(node.getChilds()).end()) {}
};

template <class T>
bool has_child (nicai::xml::element & node)
{
  return nicai::xml::poly_childs<T>(node.getChilds()).begin()
      != nicai::xml::poly_childs<T>(node.getChilds()).end();
};


std::string namespaceDeclFromFullName(std::string name, bool opening)
{
  std::string result="";
  std::string::size_type oldpos = 0;
  std::string::size_type pos = name.find("::");
  while(pos!=std::string::npos)
  {
    if (opening)
      result+="namespace "+name.substr(oldpos, pos-oldpos)+" { ";
    else
      result+="} /*namespace "+name.substr(oldpos, pos-oldpos)+"*/ ";
    oldpos=pos+2;
    pos=name.find("::", pos+2);
  }
  return result;
}

std::string nameFromFullName(std::string name)
{
  std::string::size_type oldpos = 0;
  std::string::size_type pos = name.find("::");
  while(pos!=std::string::npos)
  {
    oldpos=pos;
    pos=name.find("::", pos+1);
  }
  return name.substr(oldpos+2);
}


void switch_namespace(std::string & active_namespace, const std::string & new_namespace, std::ostream & out)
{
  std::string ons = namespaceDeclFromFullName(active_namespace, true);
  std::string nns = namespaceDeclFromFullName(new_namespace, true);
  if (ons==nns)
    return;
  out << std::endl;
  out << namespaceDeclFromFullName(active_namespace, false) << std::endl;
  out << std::endl;
  out << "////////////////////////////////////////////////////////////////////////////////" << std::endl;
  out << std::endl;
  out << nns << std::endl;
  out << std::endl;
  active_namespace=new_namespace;
}


////////////////////////////////////////////////////////////////////////////////
// generator class
////////////////////////////////////////////////////////////////////////////////


generator::~generator()
{
}


generator::generator(const std::string & basename)
: pimpl(new generator_impl())
{
  pimpl->basename=basename;
  pimpl->doc=0;
}



bool generator::readXml()
{
  std::ifstream in((pimpl->basename+".xml").c_str());
  nicai::xml::ixmlstream xin(in);
  pimpl->doc = xgegrammar::XgeDocument::trycreate(xin, true);
  return pimpl->doc;
}


bool generator::writeCode()
{
  pimpl->hpp.open((pimpl->basename+"_out.h").c_str());
  pimpl->cpp.open((pimpl->basename+"_out.cpp").c_str());

  pimpl->generate_code();

  pimpl->hpp.close();
  pimpl->cpp.close();
  return true;
}


std::string generator_impl::transform_type(const std::string & atype, const std::string & def)
{
  if (atype=="")
    return def;
  if (atype=="xsd:string")
    return "nicai::xml::xstring";
  if (atype=="xsd:integer")
    return "nicai::xml::xdecimal";
  if (atype=="xsd:boolean")
    return "nicai::xml::xboolean";
  if (atype=="xsd:float")
    return "nicai::xml::xfloat";
  if (atype=="xsd:double")
    return "nicai::xml::xdouble";
  return class_names[atype];
}

////////////////////////////////////////////////////////////////////////////////
// code generation
////////////////////////////////////////////////////////////////////////////////

void generator_impl::generate_code()
{
  using namespace xgegrammar;
  hpp << "// automatically generated code, DO NOT MODIFY!" << std::endl;
  hpp << std::endl;

  cpp << "// automatically generated code, DO NOT MODIFY!" << std::endl;
  cpp << std::endl;

  nicai::xml::poly_childs<Grammar> grammars(doc->getChilds());
  if (!grammars.empty())
    generate_code(*(grammars.begin()));
}


void generator_impl::generate_code(xgegrammar::Grammar & node)
{
  // header

  hpp << "#ifndef XGEGEN_" << node.att_name.get() << "_H" << std::endl;
  hpp << "#define XGEGEN_" << node.att_name.get() << "_H" << std::endl;
  hpp << std::endl;
  hpp << "#include \"nicai/xml/element.h\"" << std::endl;
  hpp << "#include \"nicai/xml/schematypes.h\"" << std::endl;

  generate_class_declaration(node);
  generate_class_definition(node);

  switch_namespace(active_hpp_namespace, "", hpp);
  hpp << "#endif" << std::endl;
  hpp << std::endl;

  // cpp-file

  cpp << "#include \"" << node.att_name.get() << "_out.h\"" << std::endl;
  cpp << "#include <nicai/xml/stream.h>"<< std::endl;
  cpp << std::endl;

  generate_class_implementation(node);

  switch_namespace(active_cpp_namespace, "", cpp);
}



////////////////////////////////////////////////////////////////////////////////
// class declarations
////////////////////////////////////////////////////////////////////////////////

void generator_impl::generate_class_declaration(xgegrammar::Grammar & node)
{
  for (each_child<xgegrammar::DocumentRoot> it(node); it; it++)
    generate_class_declaration(*it);
  for (each_child<xgegrammar::ComplexClass> it(node); it; it++)
    generate_class_declaration(*it);
  hpp << std::endl;
}


void generator_impl::generate_class_declaration(xgegrammar::ComplexClass & node)
{
  std::string cn = nameFromFullName(node.att_className.get());
  std::string ns = namespaceDeclFromFullName(node.att_className.get(), true);
  std::string ne = namespaceDeclFromFullName(node.att_className.get(), false);

  class_names[node.att_name] = node.att_className;

  switch_namespace(active_hpp_namespace, node.att_className.get(), hpp);
  hpp << "class " << cn << ";" << std::endl;
}


void generator_impl::generate_class_declaration(xgegrammar::DocumentRoot & node)
{
  std::string cn = nameFromFullName(node.att_className.get());
  std::string ns = namespaceDeclFromFullName(node.att_className.get(), true);
  std::string ne = namespaceDeclFromFullName(node.att_className.get(), false);

  switch_namespace(active_hpp_namespace, node.att_className.get(), hpp);
  hpp << "class " << cn << ";" << std::endl;
}


////////////////////////////////////////////////////////////////////////////////
// class definitions
////////////////////////////////////////////////////////////////////////////////

void generator_impl::generate_class_definition(xgegrammar::Grammar & node)
{
  for (each_child<xgegrammar::DocumentRoot> it(node); it; it++)
    generate_class_definition(*it);
  for (each_child<xgegrammar::ComplexClass> it(node); it; it++)
    generate_class_definition(*it);
  hpp << std::endl;
}


void generator_impl::generate_class_definition(xgegrammar::ComplexClass & node)
{
  std::string fullname = node.att_className.get();
  std::string cn = nameFromFullName(fullname);
  std::string base = transform_type(node.att_base, "nicai::xml::element");

  switch_namespace(active_hpp_namespace, fullname, hpp);
  hpp << "class " << cn << " : public " << base << std::endl;
  hpp << "{" << std::endl;
  hpp << "  public:" << std::endl;

  if (has_child<xgegrammar::Attribute>(node))
  {
    for (each_child<xgegrammar::Attribute> it(node); it; it++)
      generate_class_definition(*it);
    hpp << std::endl;
  }

  hpp << "    explicit " << cn << "(const std::string & name);" << std::endl;
  hpp << "    virtual " << cn << " * clone();" << std::endl;

  if (has_child<xgegrammar::Element>(node))
    hpp << "    virtual nicai::xml::element * createChildElement(const std::string & name, nicai::xml::ixmlbasestream & xin);" << std::endl;

  hpp << "};" << std::endl;
  hpp << std::endl;
}


void generator_impl::generate_class_definition(xgegrammar::DocumentRoot & node)
{
  std::string fullname=node.att_className.get();
  std::string cn = nameFromFullName(fullname);
  std::string base = "nicai::xml::documentroot";

  switch_namespace(active_hpp_namespace, fullname, hpp);
  hpp << "class " << cn << " : public " << base << std::endl;
  hpp << "{" << std::endl;
  hpp << "  public:" << std::endl;
  hpp << "    " << cn << "();" << std::endl;
  hpp << "    virtual " << cn << " * clone();" << std::endl;

  if (has_child<xgegrammar::Element>(node))
    hpp << "    virtual nicai::xml::element * createChildElement(const std::string & name, nicai::xml::ixmlbasestream & xin);" << std::endl;

  hpp << "    static " << cn << " * trycreate(nicai::xml::ixmlbasestream & xin, bool createContent = false);  " << std::endl;
  hpp << "};" << std::endl;
  hpp << std::endl;
}


void generator_impl::generate_class_definition(xgegrammar::Attribute & node)
{
  using namespace xgegrammar;

  std::string atype = transform_type(node.att_type);
  std::string aname = node.att_name;
  hpp << "    " << atype << " att_" << aname << ";" << std::endl;
}


////////////////////////////////////////////////////////////////////////////////
// class implementations
////////////////////////////////////////////////////////////////////////////////

void generator_impl::generate_class_implementation(xgegrammar::Grammar & node)
{
  for (each_child<xgegrammar::DocumentRoot> it(node); it; it++)
    generate_class_implementation(*it);
  for (each_child<xgegrammar::ComplexClass> it(node); it; it++)
    generate_class_implementation(*it);
  cpp << std::endl;
}

void generator_impl::generate_class_implementation(xgegrammar::ComplexClass & node)
{
  std::string fullname=node.att_className.get();
  std::string cn = nameFromFullName(fullname);
  std::string base = transform_type(node.att_base, "nicai::xml::element");

  switch_namespace(active_cpp_namespace, fullname, cpp);

  cpp << std::endl;
  cpp << "////////////////////////////////////////////////////////////////////////////////" << std::endl;
  cpp << std::endl;

  cpp << cn << "::" << cn << "(const std::string & name)" << std::endl;
  cpp << ": " << base << "(name)" << std::endl;

  if (has_child<xgegrammar::Attribute>(node))
  {
    for (each_child<xgegrammar::Attribute> it(node); it; it++)
      generate_class_implementation(*it);
  }

  cpp << "{" << std::endl;
  cpp << "}" << std::endl;
  cpp << std::endl;

  cpp << cn << " * " << cn << "::clone()" << std::endl;
  cpp << "{" << std::endl;
  cpp << "  return new " << cn << "(*this);" << std::endl;
  cpp << "}" << std::endl;
  cpp << std::endl;


  if (has_child<xgegrammar::Element>(node))
  {
    cpp << "nicai::xml::element * " << cn << "::createChildElement(const std::string & name, nicai::xml::ixmlbasestream & xin)" << std::endl;
    cpp << "{" << std::endl;

    for (each_child<xgegrammar::Element> it(node); it; it++)
      generate_class_implementation(*it);

    cpp << "  return " << base << "::createChildElement(name, xin);" << std::endl;
    cpp << "}" << std::endl;
    cpp << std::endl;
  }
}


void generator_impl::generate_class_implementation(xgegrammar::DocumentRoot & node)
{
  std::string fullname=node.att_className.get();
  std::string cn = nameFromFullName(fullname);
  std::string base = "nicai::xml::documentroot";

  switch_namespace(active_cpp_namespace, fullname, cpp);

  cpp << std::endl;
  cpp << "////////////////////////////////////////////////////////////////////////////////" << std::endl;
  cpp << std::endl;

  cpp << cn << "::" << cn << "()" << std::endl;
  cpp << "{" << std::endl;
  cpp << "}" << std::endl;
  cpp << std::endl;

  cpp << cn << " * " << cn << "::clone()" << std::endl;
  cpp << "{" << std::endl;
  cpp << "  return new " << cn << "(*this);" << std::endl;
  cpp << "}" << std::endl;
  cpp << std::endl;

  if (has_child<xgegrammar::Element>(node))
  {
    cpp << "nicai::xml::element * " << cn << "::createChildElement(const std::string & name, nicai::xml::ixmlbasestream & xin)" << std::endl;
    cpp << "{" << std::endl;

    for (each_child<xgegrammar::Element> it(node); it; it++)
      generate_class_implementation(*it);

    cpp << "  return " << base << "::createChildElement(name, xin);" << std::endl;
    cpp << "}" << std::endl;
    cpp << std::endl;
  }

  cpp << cn << " * " << cn << "::trycreate(nicai::xml::ixmlbasestream & xin, bool createContent)" << std::endl;
  cpp << "{" << std::endl;
  cpp << "  if (xin.expectNode(nicai::xml::document()))" << std::endl;
  cpp << "    return initChild(new " << cn << "(), xin, createContent);" << std::endl;
  cpp << "  return 0;" << std::endl;
  cpp << "}" << std::endl;
  cpp << std::endl;
}


void generator_impl::generate_class_implementation(xgegrammar::Attribute & node)
{
  std::string aname = node.att_name;
  cpp << ", att_" << aname << "(getAttributeRef(\"" << aname << "\"))" << std::endl;
}


void generator_impl::generate_class_implementation(xgegrammar::Element & node)
{
  std::string aname = node.att_name;
  std::string atype = transform_type(node.att_type);
  cpp << "  if (name==\"" << aname << "\") return create<" << atype << ">(name, xin, true);" << std::endl;
}


